from psycopg2 import connect as psycopg2_connect
from psycopg2.extras import RealDictCursor

from .base import EngineBase


class PostgreSQLEngine(EngineBase):
    def __init__(self, dsn=None, host=None, port=5432,
                 user=None, password=None, database=None, **kwargs):
        self._params = {
            'cursor_factory': RealDictCursor,
            **kwargs
        }
        self._dsn = dsn
        if dsn is None:
             self._dsn = "postgresql://{}:{}@{}:{}/{}".format(
                user, password, host, port, database
            )

        self._conn = self.connection()

    def connection(self):
        return psycopg2_connect(self._dsn, **self._params)

    def reset(self):
        self._conn = self.connection()

    def _fetch(self, query, params, operation):
        with self._conn.cursor() as cur:
            cur.execute(query, params)
            return getattr(cur, operation)()

    def execute(self, query, params=None):
        with self._conn.cursor() as cur:
            cur.execute(query, params)
            return {
                'lastrowid': cur.lastrowid,
                'rowcount': cur.rowcount,
                'rownumber': cur.rownumber,
            }

    def fetchall(self, query, params):
        return self._fetch(query, params, 'fetchall')

    def fetchone(self, query, params):
        return self._fetch(query, params, 'fetchone')

    def commit(self):
        self._conn.commit()

    def begin(self):
        self.execute("BEGIN")

    def rollback(self):
        self._conn.rollback()

    def mogrify(self, query, params):
        with self._conn.cursor() as cur:
            return cur.mogrify(query, params).decode()

    def close(self):
        self._conn.close()


class EngineBase:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()

    def connection(self, *args, **kwargs):
        raise NotImplementedError()

    def execute(self, query, params):
        raise NotImplementedError()

    def fetchone(self, query, params):
        raise NotImplementedError()

    def fetchall(self, query, params):
        raise NotImplementedError()

    def reset(self):
        raise NotImplementedError()

    def commit(self):
        raise NotImplementedError()

    def begin(self):
        raise NotImplementedError()

    def rollback(self):
        raise NotImplementedError()

    def mogrify(self, query, params):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


from pymysql import connect as mysql_connect
from pymysql.cursors import DictCursor


from .base import EngineBase


class MySQLEngine(EngineBase):
    def __init__(self, user=None, password=None, db=None,
                 host="localhost", port=3306, **kwargs):
        self._params = {
            'user': user,
            'password': password,
            'db': db,
            'host': host,
            'port': port,
            'cursorclass': DictCursor,
            **kwargs
        }
        self._conn = self.connection()

    def connection(self):
        return mysql_connect(**self._params)

    def reset(self):
        self._conn = self.connection()

    def _fetch(self, query, params, operation):
        with self._conn.cursor() as cur:
            cur.execute(query, params)
            return getattr(cur, operation)()

    def execute(self, query, params):
        with self._conn.cursor() as cur:
            affected_rows = cur.execute(query, params)
            return {
                'lastrowid': None,
                'rowcount': None,
                'rownumber': affected_rows
            }

    def fetchone(self, query, params):
        return self._fetch(query, params, 'fetchone')

    def fetchall(self, query, params):
        return self._fetch(query, params, 'fetchall')

    def commit(self):
        self._conn.commit()

    def begin(self):
        self._conn.begin()

    def rollback(self):
        self._conn.rollback()

    def mogrify(self, query, params):
        with self._conn.cursor() as cur:
            return cur.mogrify(query, params)

    def close(self):
        self._conn.close()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="bajra",
    version="0.0.1",
    author="Souvik Pratiher",
    author_email="spratiher9@gmail.com",
    description="A efficient raw SQL ORM",
    url="https://github.com/Spratiher9/Bajra",
    download_url="https://github.com/Spratiher9/Bajra/archive/refs/tags/v0.0.1.tar.gz",
    keywords=['sql', 'orm', 'raw-sql', 'python', 'postgres', 'mysql', 'database'],
    packages=find_packages(),
    long_description=long_description,      # Long description read from the the readme file
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],                                      # Information to filter the project on PyPi website
    python_requires='>=3.6',                # Minimum version requirement of the package
    py_modules=["bajra"],                   # Name of the python package
    install_requires=[
        "PyMySQL==0.9.2",
        "psycopg2-binary==2.9.3"
    ]                                       # Install other dependencies if any
)



class ExecuteResult:
    def __init__(self, lastrowid=None, rowcount=None, rownumber=None):
        self.lastrowid = lastrowid
        self.rowcount = rowcount
        self.rownumber = rownumber

    def __repr__(self):
        return '<ExecuteResult lastrowid={} rowcount={} rownumber={}'.format(
            self.lastrowid, self.rowcount, self.rownumber
        )


class Row:
    def __init__(self, **kwargs):
        self._data = kwargs

    def __getattr__(self, item):
        return self._data[item]

    def __getitem__(self, item):
        return self._data[item]

    def __iter__(self):
        for key, value in self._data.items():
            yield (key, value)

    def items(self):
        return self._data.items()

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def __repr__(self):
        val = ' '.join('{}={}'.format(k, v) for k, v in self.items())
        return '<Row {}>'.format(val)


class RowResult:
    def __init__(self, data, row_class=Row):
        self._data = data
        self._row_class = row_class

    def _to_row(self, data):
        return self._row_class(**data)

    def __getitem__(self, idx):
        return self._to_row(self._data[idx])

    def __iter__(self):
        for datum in self._data:
            yield self._to_row(datum)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return '<RowResult items={}>'.format(len(self))


class Query:
    def __init__(self):
        self._parts = []
        self._dict_params = {}
        self._list_params = []

    def _fusion(self, sql):
        self._parts.append(sql.as_sql())
        self._append_params(sql.get_params())

    def _append_params(self, params):
        if isinstance(params, dict):
            self._dict_params.update(params)

        elif isinstance(params, (list, tuple, set)):
            self._list_params.extend(params)

    def append(self, sql, params=None, **kwargs):
        if params is None:
            params = kwargs

        if isinstance(sql, Query):
            self._fusion(sql)
            self._append_params(params)
            return self

        self._parts.append(sql)
        self._append_params(params)
        return self

    def as_sql(self):
        return '\n'.join(self._parts)

    def get_params(self):
        if self._dict_params:
            return self._dict_params

        elif self._list_params:
            return self._list_params

        return None

    def __repr__(self):
        return self.as_sql()



from functools import wraps

from .base import Row, RowResult, Query, ExecuteResult


class Transaction:
    def __init__(self, bajra, rollback_on_exc=True, callback_exc=None):
        self._cater = bajra
        self._rollback_on_exc = rollback_on_exc
        self._callback_exc = callback_exc

    def __enter__(self):
        self._cater.begin()
        return self._cater

    def __exit__(self, exc_type, exc_value, traceback):
        returned_value = True

        if exc_type is not None and self._rollback_on_exc:
            self._cater.rollback()
            returned_value = False

        if exc_type is not None and callable(self._callback_exc):
            self._callback_exc(self._cater, exc_type, exc_value, traceback)
            returned_value = True

        self._cater.commit()
        return returned_value

    def __call__(self, fnx):
        @wraps(fnx)
        def _inner(*args, **kwargs):
            with self:
                fnx(*args, **kwargs)

        return _inner


class Bajra:
    def __init__(self, engine, autocommit=True):
        self._engine = engine
        self._autocommit = autocommit
        self._in_transaction = False

    def _commit(self):
        if self._autocommit and not self._in_transaction:
            self.commit()

    def _get_sql(self, query):
        if isinstance(query, Query):
            return query.as_sql()

        return query

    def _get_params(self, query, params):
        if not isinstance(query, Query):
            return params

        query_params = query.get_params()
        if not query_params:
            return params

        if not params:
            return query_params

        if type(query_params) != type(params):
            raise Exception("Diferent parameters format!")

        if isinstance(query_params, dict):
            query_params.update(params)
            return query_params

        query_params.extend(params)
        return query_params

    def query(self, query=None, params=None, **kwargs):
        q = Query()
        if params is None:
            params = kwargs

        if query is not None:
            q.append(query, params=params)

        return q

    def execute(self, query, params=None, **kwargs):
        if params is None:
            params = kwargs

        sql = self._get_sql(query)
        _params = self._get_params(query, params)
        data = self._engine.execute(sql, _params)
        self._commit()
        return ExecuteResult(**data)

    def fetchone(self, query, params=None, **kwargs):
        if params is None:
            params = kwargs

        sql = self._get_sql(query)
        _params = self._get_params(query, params)
        data = self._engine.fetchone(sql, _params)
        self._commit()
        if data is None:
            return None

        return Row(**data)

    def fetchall(self, query, params=None, **kwargs):
        if params is None:
            params = kwargs

        sql = self._get_sql(query)
        _params = self._get_params(query, params)
        data = self._engine.fetchall(sql, _params)
        self._commit()
        if data is None:
            return None

        return RowResult(data, row_class=Row)

    def reset(self):
        self._engine.reset()

    def commit(self):
        self._engine.commit()
        self._in_transaction = False

    def begin(self):
        self._engine.begin()
        self._in_transaction = True

    def rollback(self):
        if self._in_transaction:
            self._engine.rollback()
            self._in_transaction = False

    def transaction(self, rollback_on_exc=True, callback_exc=None):
        return Transaction(
            self, rollback_on_exc=rollback_on_exc, callback_exc=callback_exc
        )

    def mogrify(self, query, params=None, fnx=None):
        sql = self._get_sql(query)
        _params = self._get_params(query, params)
        data = self._engine.mogrify(sql, _params)
        if fnx is None:
            return data

        fnx(data)